"""
Almost a drop-in replacement to Python's builtin ``tempfile`` library. It's almosts a drop-in replacement because we
don't implemented ``SpooledTemporaryFile`` and we can't proxy the ``tempdir`` and ``template`` properties here
(no property descriptors for modules).
Instead of brewing some black magic to enable this, we chose the simple approach where if you want to change ``tempdir``
or ``template``, just import Python's ``tempfile`` and set them.
"""

from __future__ import absolute_import

import tempfile as _tempfile
from .os import _FileObjectThreadWithContext
from .deferred import create_threadpool_executed_func


__all__ = ['TemporaryFile', 'NamedTemporaryFile', 'SpooledTemporaryFile', 'mkstemp', 'mkdtemp', 'mktemp',
           'gettempdir', 'gettempprefix']


_TemporaryFile = create_threadpool_executed_func(_tempfile.TemporaryFile)


def TemporaryFile(mode="w+b", bufsize=-1, *args):
    f = _TemporaryFile(mode, bufsize, *args)
    return _FileObjectThreadWithContext(f, mode, bufsize)


_NamedTemporaryFile = create_threadpool_executed_func(_tempfile.NamedTemporaryFile)


def NamedTemporaryFile(mode="w+b", bufsize=-1, *args):
    f = _NamedTemporaryFile(mode, bufsize, *args)
    result = _FileObjectThreadWithContext(f, mode, bufsize)
    result.name = f.name
    return result


def SpooledTemporaryFile(*args):
    raise NotImplementedError()


mkstemp = create_threadpool_executed_func(_tempfile.mkstemp)
mkdtemp = create_threadpool_executed_func(_tempfile.mkdtemp)
mktemp = create_threadpool_executed_func(_tempfile.mktemp)


_gettempdir = create_threadpool_executed_func(_tempfile.gettempdir)


def gettempdir():
    if _tempfile.tempdir is None:
        return _gettempdir()
    return _tempfile.tempdir

gettempprefix = _tempfile.gettempprefix



"""
This module acts as a drop-in replacement to Python's `glob` module where the blocking operations are done in a
gevent-friendly manner.
"""
from __future__ import absolute_import
import glob as _glob
from .deferred import create_threadpool_executed_func

__all__ = ['glob', 'iglob']

glob = create_threadpool_executed_func(_glob.glob)
iglob = create_threadpool_executed_func(_glob.iglob)



import gevent
from gevent.event import Event
from infi.pyutils.contexts import contextmanager

from .safe_greenlets import uncaught_greenlet_exception_context


class GeventLoopBase(object):
    """A base class for implementing an operation that sleeps for `interval` between callback executions."""
    def __init__(self, interval):
        super(GeventLoopBase, self).__init__()
        self._interval = interval
        self._stop_event = None
        self._greenlet = None

    def _loop(self):
        """Main loop - used internally."""
        while True:
            try:
                with uncaught_greenlet_exception_context():
                    self._loop_callback()
            except gevent.GreenletExit:
                break
            if self._stop_event.wait(self._interval):
                break
        self._clear()

    def _loop_callback(self):
        """Subclasses should implement this function - called every loop iteration."""
        raise NotImplementedError()  # pragma: nocover

    def start(self):
        """
        Starts the loop. Calling a running loop is an error.
        """
        assert not self.has_started(), "called start() on an active GeventLoop"
        self._stop_event = Event()
        # note that we don't use safe_greenlets.spawn because we take care of it in _loop by ourselves
        self._greenlet = gevent.spawn(self._loop)

    def stop(self, timeout=None):
        """
        Stops a running loop and waits for it to end if timeout is set. Calling a non-running loop is an error.
        :param timeout: time (in seconds) to wait for the loop to end after signalling it. ``None`` is to block till it ends.
        :return: True if the loop stopped, False if still stopping.
        """
        assert self.has_started(), "called stop() on a non-active GeventLoop"
        greenlet = self._greenlet if gevent.getcurrent() != self._greenlet else None
        self._stop_event.set()
        if greenlet is not None and timeout is not None:
            greenlet.join(timeout)
            return greenlet.ready
        else:
            return True

    def kill(self):
        """Kills the running loop and waits till it gets killed."""
        assert self.has_started(), "called kill() on a non-active GeventLoop"
        self._stop_event.set()
        self._greenlet.kill()
        self._clear()

    def has_started(self):
        """
        :return: True if the loop is running
        """
        return self._greenlet is not None

    def _clear(self):
        self._greenlet = None
        self._stop_event = None


class GeventLoop(GeventLoopBase):
    """Loop class that expects an interval and a callback. A more usable scenario is to use `do_in_background`.

    For example:

    ```
    def do_some_period_maintenance():
        ...

    maintenance_job = GeventLoop(5, do_some_period_maintenance)
    maintenance_job.start()
    maintenance_job.stop()
    ```
    """
    def __init__(self, interval, callback):
        super(GeventLoop, self).__init__(interval)
        self._callback = callback

    def _loop_callback(self):
        self._callback()

    def __repr__(self):
        return "GeventLoop(interval={}, callback={!r})".format(self._interval, self._callback)


@contextmanager
def loop_in_background(interval, callback):
    """
    When entering the context, spawns a greenlet that sleeps for `interval` seconds between `callback` executions.
    When leaving the context stops the greenlet.
    The yielded object is the `GeventLoop` object so the loop can be stopped from within the context.

    For example:
    ```
    with loop_in_background(60.0, purge_cache) as purge_cache_job:
        ...
        ...
        if should_stop_cache():
            purge_cache_job.stop()
    ```
    """
    loop = GeventLoop(interval, callback)
    loop.start()
    try:
        yield loop
    finally:
        if loop.has_started():
            loop.stop()



import sys
import gevent


def create_threadpool_executed_func(original_func):
    """
    Returns a function wrapper that defers function calls execute inside gevent's threadpool but keeps any exception
    or backtrace in the caller's context.
    :param original_func: function to wrap
    :returns: wrapper function
    """
    def wrapped_func(*args, **kwargs):
        try:
            result = original_func(*args, **kwargs)
            return True, result
        except:
            return False, sys.exc_info()

    def new_func(*args, **kwargs):
        status, result = gevent.get_hub().threadpool.apply(wrapped_func, args, kwargs)
        if status:
            return result
        else:
            raise result[0], result[1], result[2]
    new_func.__name__ = original_func.__name__
    new_func.__doc__ = "(gevent-friendly)" + (" " + original_func.__doc__ if original_func.__doc__ is not None else "")
    return new_func



import sys
from infi.traceback import traceback_decorator
from infi.pyutils.contexts import contextmanager
import gevent
import gevent.pool
from logbook import Logger
logger = Logger(__name__)

_uncaught_exception_handler = None


class SafeGreenlet(gevent.Greenlet):
    def __init__(self, run, *args, **kwargs):
        super(SafeGreenlet, self).__init__(wrap_uncaught_greenlet_exceptions(run), *args, **kwargs)


class SafePool(gevent.pool.Pool):
    greenlet_class = SafeGreenlet


def set_greenlet_uncaught_exception_handler(func):
    """
    Sets a global greenlet uncaught exception handler that will get called if a greenlet spawned by one of this module's
    wrappers raises an uncaught exception.
    :param func: exception handler function that will receive the exc_info tuple as an argument
    :returns: previous exception handler function or None
    """
    global _uncaught_exception_handler
    prev_handler, _uncaught_exception_handler = _uncaught_exception_handler, func
    return prev_handler


@traceback_decorator
def uncaught_greenlet_exception():
    global _uncaught_exception_handler
    if _uncaught_exception_handler:
        _uncaught_exception_handler()
    else:
        logger.exception("uncaught exception in greenlet and uncaught exception handler was not set, calling sys.exit")
        sys.exit("uncaught greenlet exception")


@contextmanager
def uncaught_greenlet_exception_context():
    try:
        yield
    except gevent.GreenletExit:
        raise
    except:
        uncaught_greenlet_exception()


def _normalize_instancemethod(instance_method):
    """
    wraps(instancemethod) returns a function, not an instancemethod so its repr() is all messed up;
    we want the original repr to show up in the logs, therefore we do this trick
    """
    if not hasattr(instance_method, 'im_self'):
        return instance_method

    def _func(*args, **kwargs):
        return instance_method(*args, **kwargs)

    _func.__name__ = repr(instance_method)
    return _func


def wrap_uncaught_greenlet_exceptions(func):
    def _func(*args, **kwargs):
        with uncaught_greenlet_exception_context():
            return func(*args, **kwargs)
    _func.__name__ = repr(func)
    return _func


def safe_spawn(func, *args, **kwargs):
    return gevent.spawn(wrap_uncaught_greenlet_exceptions(func), *args, **kwargs)


def safe_spawn_later(seconds, func, *args, **kwargs):
    return gevent.spawn_later(seconds, wrap_uncaught_greenlet_exceptions(func), *args, **kwargs)


def safe_spawn_and_switch_immediately(func, *args, **kwargs):
    greenlet = safe_spawn(func, *args, **kwargs)
    gevent.sleep(0)
    return greenlet


def set_hub_error_handlers():
    hub = gevent.get_hub()
    original_hub_system_error = hub.handle_system_error

    def handle_error(context, type, value, tb):
        logger.error("encountered gevent hub error {}: {}".format(type, value))
        # the original function just logs the exceptions to stderr
        if issubclass(type, KeyboardInterrupt):
            original_hub_system_error(type, value)
        else:
            uncaught_greenlet_exception()

    def handle_system_error(type, value):
        # the original function just logs the exceptions to stderr
        logger.error("encountered gevent hub system error {}: {}".format(type, value))
        uncaught_greenlet_exception()

    gevent.get_hub().handle_error = handle_error
    gevent.get_hub().handle_system_error = handle_system_error


def safe_joinall(greenlets, timeout=None, raise_error=False):
    """
    Wrapper for gevent.joinall if the greenlet that waits for the joins is killed, it kills all the greenlets it
    joins for.
    """
    greenlets = list(greenlets)
    try:
        gevent.joinall(greenlets, timeout=timeout, raise_error=raise_error)
    except gevent.GreenletExit:
        [greenlet.kill() for greenlet in greenlets if not greenlet.ready()]
        raise
    return greenlets



"""
This module acts as a replacement to some of Python's `shutil` module where the blocking operations are done in a
gevent-friendly manner.
"""
from __future__ import absolute_import
import shutil as _shutil
from .deferred import create_threadpool_executed_func

CONSTS = ['Error']


DEFERRED_FUNCTIONS = [
    'copyfile',
    'copyfileobj',
    'copymode',
    'copystat',
    'copy',
    'copy2',
    'copytree',
    'move',
    'rmtree'
]


PASSTHROUGH_FUNCTIONS = [
    'ignore_patterns'
]


__all__ = CONSTS + DEFERRED_FUNCTIONS + PASSTHROUGH_FUNCTIONS

module = globals()
for name in DEFERRED_FUNCTIONS:
    if name in _shutil.__dict__:
        module[name] = create_threadpool_executed_func(_shutil.__dict__[name])
for name in CONSTS + PASSTHROUGH_FUNCTIONS:
    if name in _shutil.__dict__:
        module[name] = _shutil.__dict__[name]



"""
This module acts as an 99% drop-in replacement to Python's `os` module where the blocking operations are done in a
gevent-friendly manner.
When we say "blocking operations" we mean not only the obvious blockers such as read and write, but also other
operations that may trigger I/O such as close, listdir, fstat and the like.
Also, on UNIXes (and Windows through the POSIX module), read and writes from a local file may block even if we enable
O_NONBLOCK. So for them we need also need to defer to a thread.
"""
from __future__ import absolute_import
import os as _os
from gevent.queue import Queue
from gevent.hub import get_hub
from gevent.fileobject import FileObjectThread
from gevent.os import make_nonblocking, nb_read, nb_write
from gevent import fork  # Ignore lint error - it's added to __all__

from ..deferred import create_threadpool_executed_func
from . import path  # Ignore lint error -it's added to __all__

# All os.xxx attributes were extracted from Python 2.7.6 on Ubuntu
CONSTS = [
    'EX_CANTCREAT', 'EX_CONFIG', 'EX_DATAERR', 'EX_IOERR', 'EX_NOHOST', 'EX_NOINPUT', 'EX_NOPERM', 'EX_NOUSER', 'EX_OK',
    'EX_OSERR', 'EX_OSFILE', 'EX_PROTOCOL', 'EX_SOFTWARE', 'EX_TEMPFAIL', 'EX_UNAVAILABLE', 'EX_USAGE', 'F_OK',
    'NGROUPS_MAX', 'O_APPEND', 'O_ASYNC', 'O_CREAT', 'O_DIRECT', 'O_DIRECTORY', 'O_DSYNC', 'O_EXCL', 'O_LARGEFILE',
    'O_NDELAY', 'O_NOATIME', 'O_NOCTTY', 'O_NOFOLLOW', 'O_NONBLOCK', 'O_RDONLY', 'O_RDWR', 'O_RSYNC', 'O_SYNC',
    'O_TRUNC', 'O_WRONLY', 'P_NOWAIT', 'P_NOWAITO', 'P_WAIT', 'R_OK', 'SEEK_CUR', 'SEEK_END', 'SEEK_SET', 'ST_APPEND',
    'ST_MANDLOCK', 'ST_NOATIME', 'ST_NODEV', 'ST_NODIRATIME', 'ST_NOEXEC', 'ST_NOSUID', 'ST_RDONLY', 'ST_RELATIME',
    'ST_SYNCHRONOUS', 'ST_WRITE', 'TMP_MAX', 'WCONTINUED', 'WCOREDUMP', 'WEXITSTATUS', 'WIFCONTINUED', 'WIFEXITED',
    'WIFSIGNALED', 'WIFSTOPPED', 'WNOHANG', 'WSTOPSIG', 'WTERMSIG', 'WUNTRACED', 'W_OK', 'X_OK',
    'altsep', 'curdir', 'defpath', 'devnull', 'environ', 'error', 'extsep', 'linesep', 'name', 'pardir', 'pathsep',
    'sep'
]

PASSTHROUGH_FUNCTIONS = [
    '_exit', 'abort', 'confstr', 'confstr_names', 'ctermid', 'dup', 'dup2', 'execl', 'execle', 'execlp', 'execlpe',
    'execv', 'execve', 'execvp', 'execvpe', 'getcwd', 'getcwdu', 'getegid', 'getenv', 'geteuid', 'getgid', 'getgroups',
    'getloadavg', 'getlogin', 'getpgid', 'getpgrp', 'getpid', 'getppid', 'getresgid', 'getresuid', 'getsid', 'getuid',
    'initgroups', 'kill', 'killpg', 'major', 'makedev', 'minor', 'nice', 'pathconf_names', 'putenv', 'setegid',
    'seteuid', 'setgid', 'setgroups', 'setpgid', 'setpgrp', 'setregid', 'setresgid', 'setresuid', 'setreuid', 'setsid',
    'setuid', 'stat_float_times', 'strerror', 'sysconf', 'sysconf_names', 'tcgetpgrp', 'tcsetpgrp', 'times', 'ttyname',
    'umask', 'uname', 'unsetenv'
]

DEFERRED_FUNCTIONS = [
    'access', 'chdir', 'chflags', 'chmod', 'chown', 'chroot', 'close', 'closerange', 'fchdir', 'fchmod', 'fchown',
    'fdatasync', 'fpathconf', 'fstat', 'fstatvfs', 'fsync', 'ftruncate', 'isatty', 'lchflags', 'lchmod', 'lchown', 'link',
    'listdir', 'lseek', 'lstat', 'makedirs', 'mkdir', 'mkfifo', 'mknod', 'openpty', 'pathconf', 'read', 'readlink', 'remove',
    'removedirs', 'rename', 'renames', 'rmdir', 'spawnl', 'spawnle', 'spawnlp', 'spawnlpe', 'spawnv', 'spawnve',
    'spawnvp', 'spawnvpe', 'stat', 'stat_float_times', 'statvfs', 'symlink', 'system', 'tempnam', 'tmpnam', 'unlink',
    'urandom', 'utime' 'wait', 'wait3', 'wait4', 'waitpid', 'write', 'walk'
]

NOT_IMPLEMENTED_FUNCTIONS = [
    'forkpty', 'pipe', 'popen', 'popen2', 'popen3', 'popen4'
]


__all__ = ['path', 'fopen', 'open', 'read', 'write', 'fork'] + DEFERRED_FUNCTIONS + PASSTHROUGH_FUNCTIONS + CONSTS


class _FileObjectThreadWithContext(FileObjectThread):
    """Adds context manager functionality to gevent's `FileObject`."""
    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()


_fopen = create_threadpool_executed_func(open)
_fdopen = create_threadpool_executed_func(_os.fdopen)


def fopen(name, mode='r', buffering=-1):
    """Similar to Python's built-in `open()` function."""
    f = _fopen(name, mode, buffering)
    return _FileObjectThreadWithContext(f, mode, buffering)
fopen.__doc__ = open.__doc__


def fdopen(fd, mode='r', buffering=-1):
    f = _fdopen(fd, mode, buffering)
    return _FileObjectThreadWithContext(f, mode, buffering)
fdopen.__doc__ = _os.fdopen.__doc__


_tmpfile = create_threadpool_executed_func(_os.tmpfile)


def tmpfile():
    f = _tmpfile()
    return _FileObjectThreadWithContext(f, "w+b")
tmpfile.__doc__ = _os.tmpfile.__doc__


_open = create_threadpool_executed_func(_os.open)


def open(file, flags, mode=0777):
    fd = _open(file, flags, mode)
    # this is (almost) pointless since local files may block even if we set non-blocking on Linux
    # it's still interesting if the file is a named pipe, UNIX-domain socket, etc.
    make_nonblocking(fd)
    return fd


# Here we convert most of os's file-related functions to functions that are deferred to a thread so they'll be
# gevent-friendly: they still will block, but will release the greenlet so other greenlets can execute.
module = globals()
for name in DEFERRED_FUNCTIONS:
    if name in _os.__dict__:
        module[name] = create_threadpool_executed_func(_os.__dict__[name])
for name in CONSTS + PASSTHROUGH_FUNCTIONS:
    if name in _os.__dict__:
        module[name] = _os.__dict__[name]
for name in NOT_IMPLEMENTED_FUNCTIONS:
    if name in _os.__dict__:
        def _not_implemented(*args, **kwargs):
            raise NotImplementedError()
        _not_implemented.__name__ = name
        module[name] = _not_implemented



"""
greenlets that don't print exceptions to the screen and can reraise the exception through 'join' / 'joinall'
"""

import gevent
import sys
from logging import getLogger

logger = getLogger(__name__)


class SilentGreenletExceptionWrapper(object):
    def __init__(self, exc_info):
        self.exc_info = exc_info

    def get_exc_info(self):
        return self.exc_info


def wrap_uncaught_greenlet_exceptions(func):
    def _func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            logger.exception("uncaught exception in greenlet")
            return SilentGreenletExceptionWrapper(sys.exc_info())
    _func.__name__ = repr(func)
    return _func


def spawn(func, *args, **kwargs):
    """ spawns a greenlet that does not print exceptions to the screen.
    if you use this function you MUST use this module's join or joinall otherwise the exception will be lost """
    return gevent.spawn(wrap_uncaught_greenlet_exceptions(func), *args, **kwargs)


def join(greenlet, timeout=None, raise_error=True):
    value = greenlet.get(block=True, timeout=timeout)
    if isinstance(value, SilentGreenletExceptionWrapper):
        if raise_error:
            raise value.get_exc_info()
        else:
            return None
    return value


def joinall(greenlets, timeout=None, raise_error=True):
    # this will not raise or print errors if the greenlets were spawned with 'spawn' from this module:
    gevent.joinall(greenlets, timeout=timeout)
    # raise errors if needed and there were any
    if raise_error:
        [join(greenlet) for greenlet in greenlets]



from infi.pyutils.lazy import cached_function

import json
import cjson
import time
import gevent
from StringIO import StringIO

import logbook
logger = logbook.Logger(__name__, level=logbook.CRITICAL)  # by default don't log errors


class DecodeError(Exception):
    pass


class EncodeError(Exception):
    pass


class GreenletFriendlyStringIO(StringIO):
    """ A StringIO subclass that yields to other greenlets every ~0.01 secs """
    def __init__(self):
        StringIO.__init__(self)  # can't use super() since StringIO is an old-style class
        self.last_sleep = 0

    def write(self, s):
        StringIO.write(self, s)
        t = time.time()
        if t - self.last_sleep > 0.01:
            gevent.sleep(0)
            self.last_sleep = t


@cached_function
def can_dumps_sort_keys():
    from platform import python_version
    from pkg_resources import parse_version
    return parse_version(python_version()) >= parse_version("2.7.5")


def encode(python_object, indent=None, large_object=False):
    """:returns: a JSON-representation of the object"""
    # ujson and cjson say they are considerably faster than the json module in the standard library
    # so we wanted to use those where possible
    # but, ujson.encode causes segmentation fault on some non-builtin objects
    # and we get cjson.EncodeError quite a lot
    # so we're just going to use the standard json module

    # sorted keys is eaiser to read; however, Python-2.7.2 does not have this feature
    kwargs = dict(indent=indent)
    if can_dumps_sort_keys():
        kwargs.update(sort_keys=True)
    try:
        if large_object:
            out = GreenletFriendlyStringIO()
            json.dump(python_object, out, **kwargs)
            return out.getvalue()
        else:
            return json.dumps(python_object, **kwargs)
    except Exception:
        logger.exception()
        raise EncodeError('Cannot encode {!r}', python_object)


def decode(json_string):
    """:returns: a Python object"""
    # unlike in encode, we haven't seen errors in decoding, so we will use cjson as a first option, since it is faster.
    # we don't try ujson since it gets stuck on very large integers such as 2**65
    try:
        return cjson.decode(json_string)
    except cjson.DecodeError:
        logger.warning("failed to decode {} via cjson", json_string)
        try:
            return json.loads(json_string)
        except:
            raise DecodeError()



"""
This module acts as a replacement to some of Python's `os.path` module where the blocking operations are done in a
gevent-friendly manner.
"""
from __future__ import absolute_import
from os import path as _path
from ..deferred import create_threadpool_executed_func

CONSTS = ['altsep', 'extsep', 'pardir', 'pathsep', 'sep', 'supports_unicode_filenames']


DEFERRED_FUNCTIONS = [
    'abspath', 'exists', 'getatime', 'getctime', 'getmtime', 'getsize', 'isdir', 'isfile', 'islink', 'ismount',
    'lexists', 'realpath', 'samefile', 'sameopenfile', 'samestat', 'walk'
]

PASSTHROUGH_FUNCTIONS = [
    'basename', 'commonprefix', 'dirname', 'expanduser', 'expandvars', 'isabs', 'join', 'normcase', 'normpath', 'split',
    'splitdrive', 'splitext', 'relpath'
]


module = globals()
for name in DEFERRED_FUNCTIONS:
    if name in _path.__dict__:
        module[name] = create_threadpool_executed_func(_path.__dict__[name])
for name in CONSTS + PASSTHROUGH_FUNCTIONS:
    if name in _path.__dict__:
        module[name] = _path.__dict__[name]
